from __future__ import annotations

import gc
import logging
import math
import warnings
from collections.abc import Iterable
from typing import Literal

import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl
import polars._typing
import scipy.stats
import tqdm

logger = logging.getLogger(__name__)

ACTIVITY_DRIFT_THRESHOLD = 0.1


def insert_is_observed(
    intervals_frame: polars._typing.FrameType,
    units_frame: polars._typing.FrameType,
    col_name: str = "is_observed",
    unit_id_col: str = "unit_id",
) -> polars._typing.FrameType:

    if isinstance(intervals_frame, pl.LazyFrame):
        intervals_lf = intervals_frame
    else:
        intervals_lf = intervals_frame.lazy()

    if isinstance(units_frame, pl.LazyFrame):
        units_lf = units_frame
    else:
        units_lf = units_frame.lazy()

    units_schema = units_lf.collect_schema()
    if unit_id_col not in units_schema:
        raise ValueError(
            f"units_frame does not contain {unit_id_col!r} column: can be customized by passing unit_id_col"
        )
    if "obs_intervals" not in units_schema:
        raise ValueError("units_frame must contain 'obs_intervals' column")

    unit_ids = units_lf.select(unit_id_col).collect().get_column(unit_id_col).unique()
    intervals_schema = intervals_lf.collect_schema()
    if unit_id_col not in intervals_schema:
        if len(unit_ids) > 1:
            raise ValueError(
                f"units_frame contains multiple units, but intervals_frame does not contain {unit_id_col!r} column to perform join"
            )
        elif len(unit_ids) == 0:
            raise ValueError(
                f"units_frame contains no unit ids in {unit_id_col=} column"
            )
        else:
            intervals_lf = intervals_lf.with_columns(
                pl.lit(unit_ids[0]).alias(unit_id_col)
            )
    if not all(c in intervals_schema for c in ("start_time", "stop_time")):
        raise ValueError(
            "intervals_frame must contain 'start_time' and 'stop_time' columns"
        )

    if units_schema["obs_intervals"] in (
        pl.List(pl.List(pl.Float64())),
        pl.List(pl.List(pl.Int64())),
        pl.List(pl.List(pl.Null())),
    ):
        logger.info("Converting 'obs_intervals' column to list of lists")
        units_lf = units_lf.explode("obs_intervals")
    assert (type_ := units_lf.collect_schema()["obs_intervals"]) == pl.List(
        pl.Float64
    ), f"Expected exploded obs_intervals to be pl.List(f64), got {type_}"
    intervals_lf = (
        intervals_lf.join(
            units_lf.select(unit_id_col, "obs_intervals"), on=unit_id_col, how="left"
        )
        .with_columns(
            pl.when(
                pl.col("obs_intervals").list.get(0).gt(pl.col("start_time"))
                | pl.col("obs_intervals").list.get(1).lt(pl.col("stop_time")),
            )
            .then(pl.lit(False))
            .otherwise(pl.lit(True))
            .alias(col_name),
        )
        .group_by("unit_id", "start_time")
        .agg(
            pl.all().exclude("obs_intervals", col_name).first(),
            pl.col(col_name).any(),
        )
    )
    if isinstance(intervals_frame, pl.LazyFrame):
        return intervals_lf
    return intervals_lf.collect()


def get_per_trial_spike_times(
    units_df: pl.DataFrame,
    trials_df: pl.DataFrame,
    starts: pl.Expr | Iterable[pl.Expr],
    ends: pl.Expr | Iterable[pl.Expr],
    col_names: str | Iterable[str],
    apply_obs_intervals: bool = True,
    as_counts: bool = False,
) -> pl.DataFrame:
    """"""
    if isinstance(starts, pl.Expr):
        starts = (starts,)
    if isinstance(ends, pl.Expr):
        ends = (ends,)
    if isinstance(col_names, str):
        col_names = (col_names,)
    assert isinstance(col_names, tuple)  # for mypy
    assert isinstance(starts, tuple)
    assert isinstance(ends, tuple)
    if len(set(col_names)) != len(col_names):
        raise ValueError("col_names must be unique")
    if len(starts) != len(ends) != len(col_names):
        raise ValueError("starts, ends, and col_names must have the same length")

    # temp add columns for each interval with type list[float] (start, end)
    temp_col_prefix = "__temp_interval"
    for start, end, col_name in zip(starts, ends, col_names):
        trials_df = trials_df.with_columns(
            pl.concat_list(start, end).alias(f"{temp_col_prefix}_{col_name}"),
        )

    results: dict[str, list] = {
        "unit_id": [],
        "trial_index": trials_df["trial_index"].to_list() * len(units_df),
    }
    for col_name in col_names:
        results[col_name] = []
    for row in units_df.select("unit_id", "spike_times").iter_rows(named=True):
        if row["unit_id"] is None:
            raise ValueError(f"Missing unit_id in {row=}")
        results["unit_id"].extend([row["unit_id"]] * len(trials_df))

        for start, end, col_name in zip(starts, ends, col_names):
            # get spike times with start:end interval for each row of the trials table
            spike_times = row["spike_times"]
            spikes_in_all_intervals: list[list[float | int] | float | int] = []
            for a, b in np.searchsorted(
                spike_times,
                trials_df[f"{temp_col_prefix}_{col_name}"].to_list(),
            ):
                interval_spike_times: npt.NDArray[np.floating] = spike_times[a:b]
                #! spikes coincident with end of interval are not included
                if as_counts:
                    spikes_in_all_intervals.append(len(interval_spike_times))
                else:
                    spikes_in_all_intervals.append(interval_spike_times.tolist())
            results[col_name].extend(spikes_in_all_intervals)

    if apply_obs_intervals:
        results_df = trials_df.drop(pl.selectors.starts_with(temp_col_prefix)).join(
            other=pl.DataFrame(results),
            on="trial_index",
            how="left",
        )
    else:
        results_df = pl.DataFrame(results)

    if apply_obs_intervals:
        results_df = insert_is_observed(
            intervals_frame=results_df,
            units_frame=units_df,
        ).with_columns(
            *[
                pl.when(pl.col("is_observed").not_())
                .then(pl.lit(float("nan")))
                .otherwise(pl.col(col_name))
                .alias(col_name)
                for col_name in col_names
            ]
        )
    return results_df


def get_test_samples(
    unit_df: pl.DataFrame, interval: Literal["baseline", "response", "trial"]
) -> list[list[int]] | None:
    if interval == "trial" and "trial" not in unit_df.columns:
        unit_df = unit_df.with_columns(trial=pl.col("baseline") + pl.col("response"))
    if unit_df[interval].drop_nans().sum() == 0:
        # cannot perform test if all samples are the same (zero spikes is a common case, or all nans
        # when no trials observed)
        return None
    if unit_df.n_unique("block_index") < 2:
        # for Templeton we chunk spike counts into 3 segments of time
        return [
            s[interval].to_list()
            for s in unit_df.sort("trial_index").iter_slices(
                math.ceil(len(unit_df) / 3)
            )
        ]
    return (
        unit_df.group_by("block_index")
        .agg(pl.col(interval))
        .get_column(interval)
        .to_list()
    )


def add_activity_drift_metric(
    units_df: pd.DataFrame,
    trials_df: pd.DataFrame,
    col_name: str = "activity_drift",
) -> pd.DataFrame:

    units_pl = pl.from_pandas(units_df[["unit_id", "spike_times", "obs_intervals"]])
    trials_pl = pl.from_pandas(trials_df[:])
    trials_with_counts = get_per_trial_spike_times(
        units_df=units_pl,
        trials_df=trials_pl,
        starts=(
            pl.col("stim_start_time") - 1.5,
            pl.col("stim_start_time"),
        ),
        ends=(
            pl.col("stim_start_time"),
            pl.col("stim_start_time") + 2,
        ),
        col_names=(
            "baseline",
            "response",
        ),
        as_counts=True,
    )
    del units_pl
    del trials_pl
    gc.collect()

    class NullResult:
        statistic = np.nan
        pvalue = np.nan

    null_result = NullResult()

    test_results = []
    with np.errstate(over="ignore"), warnings.catch_warnings():
        warnings.simplefilter("ignore")

        for (unit_id, rewarded_modality, stim_name, *_), unit_df in tqdm.tqdm(
            iterable=(
                trials_with_counts.drop_nulls(["baseline", "response"])
                .filter(pl.col("stim_name") != "catch")
                .group_by(
                    "unit_id", "rewarded_modality", "stim_name", maintain_order=True
                )
            ),
            total=trials_with_counts.n_unique("unit_id"),
            unit="unit",
            desc="calculating activity drift",
            ncols=80,
        ):
            result = dict(
                unit_id=unit_id,
                rewarded_modality=rewarded_modality,
                stim_name=stim_name,
            )
            for interval in (
                "trial",
            ):  # previous intervals we explored were "baseline" and "response"
                samples = get_test_samples(unit_df, interval)
                if samples is None:
                    stats = null_result
                else:
                    try:
                        stats = scipy.stats.anderson_ksamp(
                            samples,
                            midrank=False,
                            method=None,
                            # even with 10k permutations, majority of p-values are clipped at 0.0001, which is useless
                            # so use default lookup table method, which is fast, then discard p-value
                        )
                    except RuntimeWarning:
                        logger.warning(
                            f"Warning encountered calculating AD test for {unit_id}, {interval=}: activity_drift will be set to NaN"
                        )
                        stats = null_result
                    except ValueError as e:
                        if "distinct" in e.args[0]:
                            logger.warning(
                                f"Distinct sample error encountered calculating AD test for {unit_id}, {interval=}: activity_drift will be set to NaN"
                            )
                            stats = null_result
                        else:
                            raise
                result[f"ad_stat_{interval}"] = stats.statistic
            test_results.append(result)
    max_min_df = (
        pl.DataFrame(test_results)
        .select(
            "unit_id",
            pl.col("ad_stat_trial")
            .max()
            .over("unit_id")
            .truediv(100)
            .clip(0, 1)
            .alias(col_name),
            # very few units have slightly negative scores, so we clip them to 0
        )
        .with_columns(
            pl.col(col_name).lt(ACTIVITY_DRIFT_THRESHOLD).alias("is_not_drift")
        )
        .unique("unit_id")
    )
    return units_df.merge(max_min_df.to_pandas(), on="unit_id", how="left")



from npc_sessions.utils.electrodes import *
from npc_sessions.utils.intervals import *
from npc_sessions.utils.misc import *
from npc_sessions.utils.units import *
from npc_sessions.utils.videos import *



from __future__ import annotations

import tempfile
from typing import Literal

import ndx_pose
import npc_lims
import npc_mvr
import numpy as np
import numpy.typing as npt
import pandas as pd
import upath
import zarr

# nice little trick from carter peene
MODEL_FUNCTION_MAPPING = {
    "dlc_eye": npc_lims.get_dlc_eye_s3_paths,
    "dlc_side": npc_lims.get_dlc_side_s3_paths,
    "dlc_face": npc_lims.get_dlc_face_s3_paths,
}

FACEMAP_CAMERA_NAMES: tuple[npc_mvr.CameraName, ...] = ("behavior", "face")
LP_CAMERA_NAMES = ("side", "face")
LP_MAPPING = {"behavior": "side", "face": "face"}
LP_RESULT_TYPES = ("predictions", "error", "temporal_norm")
LP_RESULT_DESCRIPTIONS = {
    "error": "PCA error. Pose PCA loss is the pixel error between the original pose prediction and the reconstruction of the same pose from a learnt low-dimensional representation of the body parts",
    "temporal_norm": "temporal norm. Temporal difference loss for each body part is the Euclidean distance between consecutive predictions in pixels",
}
LP_VIDEO_FEATURES_MAPPING = {
    "side": (
        "eye_top_l",
        "eye_bottom_l",
        "whisker_pad_l_top",
        "whisker_pad_l_side",
        "ear_tip_l",
        "ear_base_l",
        "nostril_l",
        "nose_tip",
        "jaw",
        "tongue_base_l",
        "tongue_tip",
    ),
    "face": (
        "eye_top_l",
        "eye_bottom_l",
        "whisker_pad_l_top",
        "whisker_pad_l_side",
        "ear_tip_l",
        "ear_base_l",
        "nostril_l",
        "nostril_r",
        "nose_tip",
        "tongue_base_l",
        "tongue_tip",
        "tongue_base_r",
    ),
}


def get_dlc_session_paf_graph(session: str, model_name: str) -> list:
    """
    https://github.com/DeepLabCut/DLC2NWB/blob/main/dlc2nwb/utils.py#L139
    >>> get_dlc_session_paf_graph('676909_2023-12-12', 'dlc_eye')
    []
    """
    model_s3_paths = MODEL_FUNCTION_MAPPING[model_name](session)
    metadata_pickle_file = tuple(
        path for path in model_s3_paths if path.suffix == ".pickle"
    )
    if not metadata_pickle_file:
        raise FileNotFoundError(
            f"No metadata pickle found for session {session}. Check {model_name} capsule"
        )

    metadata_pickle = pd.read_pickle(metadata_pickle_file[0])
    test_config = metadata_pickle["data"]["DLC-model-config file"]
    paf_graph = test_config.get("partaffinityfield_graph", [])
    if paf_graph:
        paf_inds = test_config.get("paf_best")
        if paf_inds is not None:
            paf_graph = [paf_graph[i] for i in paf_inds]

    return paf_graph


def h5_to_dataframe(h5_path: upath.UPath, key_name: str | None = None) -> pd.DataFrame:
    with tempfile.TemporaryDirectory() as tempdir:
        with open(f"{tempdir}/{h5_path.stem}", "wb") as h5_file:
            h5_file.write(h5_path.read_bytes())
            if key_name is not None:
                df_h5 = pd.read_hdf(f"{tempdir}/{h5_path.stem}", key=key_name)
            else:
                df_h5 = pd.read_hdf(f"{tempdir}/{h5_path.stem}")

    return df_h5


def _get_LPFaceParts_predictions_dataframe(
    df_predictions: pd.DataFrame,
) -> pd.DataFrame:
    df_predictions.columns = df_predictions.columns.droplevel(0)
    data_array = df_predictions.to_numpy()
    col_names = []
    for col_prefix, col_suffix in df_predictions.columns:
        col_name = f"{col_prefix}_{col_suffix}"
        col_names.append(col_name)
    return pd.DataFrame(data_array, columns=col_names).drop(
        columns="bodyparts_coords", errors="ignore"
    )


def get_LPFaceParts_result_dataframe(
    session: str, camera: str, result_name: str
) -> pd.DataFrame:
    """
    Gets the result dataframe with the lightning pose prediction for the facial features for the given camera.
    Gets one of predictions, pca singleview, or temporal norm
    Modifies the result dataframe for easier use depending on result name
    >>> df_predictions = get_LPFaceParts_result_dataframe('702136_2024-03-07', 'side', 'predictions')
    >>> len(df_predictions.columns)
    33
    >>> df_predictions.columns
    Index(['eye_top_l_x', 'eye_top_l_y', 'eye_top_l_likelihood', 'eye_bottom_l_x',
           'eye_bottom_l_y', 'eye_bottom_l_likelihood', 'whisker_pad_l_top_x',
           'whisker_pad_l_top_y', 'whisker_pad_l_top_likelihood',
           'whisker_pad_l_side_x', 'whisker_pad_l_side_y',
           'whisker_pad_l_side_likelihood', 'ear_tip_l_x', 'ear_tip_l_y',
           'ear_tip_l_likelihood', 'ear_base_l_x', 'ear_base_l_y',
           'ear_base_l_likelihood', 'nostril_l_x', 'nostril_l_y',
           'nostril_l_likelihood', 'nose_tip_x', 'nose_tip_y',
           'nose_tip_likelihood', 'jaw_x', 'jaw_y', 'jaw_likelihood',
           'tongue_base_l_x', 'tongue_base_l_y', 'tongue_base_l_likelihood',
           'tongue_tip_x', 'tongue_tip_y', 'tongue_tip_likelihood'],
          dtype='object')
    """
    if camera not in LP_CAMERA_NAMES:
        raise ValueError(
            f"Undefined camera name {camera}. Must be one of either side or face"
        )

    if result_name not in LP_RESULT_TYPES:
        raise ValueError(
            f"Undefined output type {result_name}. Must be one of either predictions, error, or norm"
        )

    session_LP_predictions_camera_s3_paths = (
        npc_lims.get_lpfaceparts_camera_predictions_s3_paths(session, camera)
    )
    session_LP_result_csv_s3_path = next(
        (
            path
            for path in session_LP_predictions_camera_s3_paths
            if f"_{result_name}.csv" in str(path)
        ),
        None,
    )
    if not session_LP_result_csv_s3_path:
        raise FileNotFoundError(
            f"{session} has no lightning pose {result_name} csv in result. Check codeocean"
        )
    headers = {"predictions": [0, 1, 2], "error": [0], "temporal_norm": [0]}
    df = pd.read_csv(session_LP_result_csv_s3_path, header=headers[result_name])
    if result_name == "predictions":
        return _get_LPFaceParts_predictions_dataframe(df).drop(
            columns="Unnamed: 0", errors="ignore"
        )
    else:
        return df.drop(columns="Unnamed: 0", errors="ignore")


def get_ellipse_session_dataframe_from_h5(session: str) -> pd.DataFrame:
    """
    >>> df_ellipse = get_ellipse_session_dataframe_from_h5('676909_2023-12-12')
    >>> len(df_ellipse)
    512347
    """
    eye_s3_paths = npc_lims.get_dlc_eye_s3_paths(session)
    ellipse_h5_path = tuple(
        path for path in eye_s3_paths if path.stem == "ellipses_processed"
    )
    if not ellipse_h5_path:
        raise FileNotFoundError(
            f"No ellipse h5 file found for {session}. Check dlc eye capsule"
        )

    # verbatim from allensdk: allensdk.brain_observatory.behavior.eye_tracking_processing.py
    eye_tracking_fields = ["cr", "eye", "pupil"]

    eye_tracking_dfs = []
    for field_name in eye_tracking_fields:
        df_ellipse = h5_to_dataframe(ellipse_h5_path[0], key_name=field_name)
        new_col_name_map = {
            col_name: f"{field_name}_{col_name}"
            for col_name in df_ellipse.columns
            if "center" in col_name
            or "width" in col_name
            or "height" in col_name
            or "phi" in col_name
        }
        df_ellipse.rename(new_col_name_map, axis=1, inplace=True)
        eye_tracking_dfs.append(df_ellipse)

    eye_tracking_data = pd.concat(eye_tracking_dfs, axis=1)
    eye_tracking_data.index.name = "frame"

    # Values in the hdf5 may be complex (likely an artifact of the ellipse
    # fitting process). Take only the real component.
    eye_tracking_data = eye_tracking_data.apply(lambda x: np.real(x.to_numpy()))

    return eye_tracking_data


def get_dlc_session_model_dataframe_from_h5(
    session: str, model_name: str
) -> pd.DataFrame:
    """
    >>> df_model = get_dlc_session_model_dataframe_from_h5('676909_2023-12-12', 'dlc_eye')
    >>> len(df_model)
    512347
    """
    model_s3_paths = MODEL_FUNCTION_MAPPING[model_name](session)
    h5_files = tuple(
        path
        for path in model_s3_paths
        if path.suffix == ".h5" and "ellipses" not in path.stem
    )

    if not h5_files:
        raise FileNotFoundError(
            f"No h5 files found for {session}. Check {model_name} capsule"
        )

    df_model = h5_to_dataframe(h5_files[0])
    return df_model


def get_pose_series_from_dataframe(
    session: str, df: pd.DataFrame, video_timestamps: npt.NDArray[np.float64]
) -> list[ndx_pose.pose.PoseEstimationSeries]:
    # https://github.com/DeepLabCut/DLC2NWB/blob/main/dlc2nwb/utils.py#L189
    if df.shape[0] != len(video_timestamps):
        raise ValueError(
            f"{session} pose dataframe has {df.shape[0]} rows, but {len(video_timestamps)} timestamps were provided."
            "\nMake sure another data asset wasn't attached to the capsule when run."
        )
    pose_estimations_series = []
    for keypoint, xy_positions in df.T.groupby(level="bodyparts", sort=False):
        assert xy_positions.shape[0] == 3
        data = xy_positions.to_numpy().T
        pose_estimation_series = ndx_pose.pose.PoseEstimationSeries(
            name=keypoint,
            description=f"{keypoint} keypoint position in each frame",  # TODO use a lookup table of abbreviation: description
            data=data[:, :2],
            unit="pixels",
            reference_frame="(0,0) corresponds to the top left corner of the video frame",
            timestamps=video_timestamps,
            confidence=data[:, 2],
            confidence_definition="softmax output of the deep neural network",
        )
        pose_estimations_series.append(pose_estimation_series)

    return pose_estimations_series


def get_facemap_output_from_s3(
    session: str, camera_name: npc_mvr.CameraName, array_name: Literal["motSVD", "proc"]
) -> zarr.Array:  # currently only saving motSVD
    """
    >>> behavior_motion_svd = get_facemap_output_from_s3('646318_2023-01-17', 'Behavior', 'motSVD')
    >>> behavior_motion_svd.shape
    (284190, 500)
    """
    session_facemap_paths = npc_lims.get_facemap_s3_paths(session)
    if not session_facemap_paths:
        raise FileNotFoundError(
            f"No facemap files found for session {session}. Check codeocean"
        )
    camera_name = npc_mvr.get_camera_name(camera_name)
    if camera_name not in FACEMAP_CAMERA_NAMES:
        raise ValueError(f"{camera_name} not part of facemap output")
    try:
        file_path = next(
            path
            for path in session_facemap_paths
            if camera_name in path.stem.lower()
            and array_name.lower() in path.stem.lower()
        )
    except StopIteration:
        raise FileNotFoundError(
            f"No {array_name} file found for session {session} {camera_name} camera, despite other paths existing. Check codeocean"
        ) from None
    else:
        return zarr.open(file_path)


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from typing import Optional

import npc_ephys
import npc_io
import npc_lims
import npc_mvr
import npc_samstim
import npc_session
import npc_stim
import npc_sync

import npc_sessions.utils as utils
from npc_sessions.trials.property_dict import PropertyDict
from npc_sessions.trials.TaskControl import TaskControl
from npc_sessions.trials.TaskControl.DynamicRouting1 import DynamicRouting1
from npc_sessions.trials.TaskControl.LuminanceTest import LuminanceTest
from npc_sessions.trials.TaskControl.OptoTagging import OptoTagging
from npc_sessions.trials.TaskControl.RFMapping import AudRFMapping, VisRFMapping


def get_trials(
    *stim_path: npc_io.PathLike,
    sync_path_or_data: Optional[npc_sync.SyncPathOrDataset],
    **kwargs,
) -> tuple[TaskControl, ...]:
    """Get trials for a given stimulus file.

    Parameters
    ----------
    stim_path : npc_io.PathLike
        Path to stimulus file.

    Returns
    -------
    PropertyDict
        Trials for the given stimulus file.
    """
    trials: list[TaskControl] = []
    for stim in stim_path:
        stim_name = npc_io.from_pathlike(stim).name
        if "DynamicRouting1" in stim_name:
            trials.append(DynamicRouting1(stim, sync_path_or_data, **kwargs))
        elif "OptoTagging" in stim_name:
            trials.append(OptoTagging(stim, sync_path_or_data, **kwargs))
        elif "RFMapping" in stim_name:
            trials.append(AudRFMapping(stim, sync_path_or_data, **kwargs))
            trials.append(VisRFMapping(stim, sync_path_or_data, **kwargs))
        elif "LuminanceTest" in stim_name:
            trials.append(LuminanceTest(stim, sync_path_or_data, **kwargs))
        else:
            raise NotImplementedError(f"Stimulus file {stim_name} not supported.")
    return tuple(trials)


__all__ = [
    "PropertyDict",
    "TaskControl",
    "OptoTagging",
    "AudRFMapping",
    "VisRFMapping",
    "DynamicRouting1",
    "LuminanceTest",
    "get_trials",
]
